package com.fundi.framework.common.comparator;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Collections;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import com.fundi.framework.common.displaylist.DisplayList;
import com.fundi.framework.common.displaylist.ListField;
import com.fundi.framework.common.displaylist.ListFieldDefn;
import com.fundi.framework.common.displaylist.ListFilter;
import com.fundi.framework.common.displaylist.ListRow;
import com.fundi.framework.common.instance.AppInstance;
import com.fundi.framework.common.nl1.Messages;
import com.fundi.framework.common.util.NumberEditor;

public class ComparatorHelper {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	private AppInstance instance;
	
	private ArrayList<Integer> sourceIds = new ArrayList<Integer>();
	private ArrayList<String> keyFields = new ArrayList<String>();
	private ArrayList<String> ignoreFields = new ArrayList<String>();
	private ArrayList<String> hideFields = new ArrayList<String>();
	private String[] labels;
	private String[][] sourceFields;
	private boolean showAllKeys = true;
	private boolean showMatchedKeys = false;
	private boolean showUnmatchedKeys = false;
	
	public static final String comparatorSavedDisplayPfx = "Comparator_"; //$NON-NLS-1$
	public static final String savedConfigurationPfx = "FUN_CompareConfig_"; //$NON-NLS-1$
	public static final String savedConfigurationSfx = ".xml"; //$NON-NLS-1$
	
	public ComparatorHelper(AppInstance instance, int origSourceId) {

		super();
		
		this.instance = instance;
		if (origSourceId >= 0) {  //Ignore it if it isn't a real comparator source (opened comparator from an empty list).
			addSource(instance.getComparatorSource(origSourceId));
		}
		
		instance.retrieveSavedDisplayComparatorSources();
		
		if (origSourceId > -1) { //Can only restore the configuration if we aren't restoring a compare from the prior session.
			restoreSavedConfiguration();
		}
	}
	
	public DisplayList compare() {
		
		DisplayList list = new DisplayList(instance.getInstance());
		int idxSce;
		int idxRow;
		int idxFld;
		DisplayList origList;
		ArrayList<ListRow> rows;
		ListRow row;
		ArrayList<ListFieldDefn> defns;
		ListFieldDefn defn;
		String keys = ""; //$NON-NLS-1$
		int idxLbl;
		ArrayList<ListField> fields;
		ListField field;
		boolean fieldSameInAllRows;
		boolean fieldContainsDifference;
		String prevValue = ""; //$NON-NLS-1$
		String value;
		
		list.setCompareRows(true);
		list.setShowAllKeys(showAllKeys);
		list.setShowMatchedKeys(showMatchedKeys);
		list.setShowUnmatchedKeys(showUnmatchedKeys);
		
		//For each list, add each visible field's definition, then add the rows.  Note that a deep copy of list fields, definitions and rows is needed.
		for (idxSce = 0; idxSce < sourceIds.size(); idxSce++) {
			//Add those definitions corresponding to a configured label.
			origList = (DisplayList) instance.getComparatorSource(sourceIds.get(idxSce)).getData(instance);
			defns = (ArrayList<ListFieldDefn>) origList.getFieldDefns().clone();
			for (idxFld = 0; idxFld < defns.size(); idxFld++) {
				defn = new ListFieldDefn(defns.get(idxFld));
				for (idxLbl = 0; idxLbl < labels.length; idxLbl++) {
					if (sourceFields[idxLbl][idxSce].equals(defn.getLabel())) {
						if (!hideFields.contains(labels[idxLbl])) {
							defn.setCompareLabel(labels[idxLbl]);
							defn.setKey(false);
							list.addFieldDefn(defn);
						}
					}
				}
			}
			
			rows = (ArrayList<ListRow>) origList.retrieveVisibleRows().clone();
			for (idxRow = 0; idxRow < origList.getVisibleRowCount(); idxRow++) {
				row = new ListRow();
				fields = rows.get(idxRow).getFields();
				for (idxFld = 0; idxFld < fields.size(); idxFld++) {
					field = fields.get(idxFld);
					row.addField(new ListField(new ListFieldDefn(field.getDefn()), field));
				}
				list.addRow(row, Integer.toString(idxSce + 1));
			}
		}
		
		for (idxFld = 0; idxFld < keyFields.size(); idxFld++) {
			if (idxFld > 0) {
				keys = keys + ";"; //$NON-NLS-1$
			}
			keys = keys + keyFields.get(idxFld);
		}
		list.setKeyFields(keys);
		
		list.setIgnoreFields(ignoreFields);
		list.setHideFields(hideFields);
		list.setComparatorSourceIds(sourceIds);
		
		if (getCurrentSources().get(0).isInstantComparison()) {
			//Hide fields which don't contain a difference, or which are ignore fields.  Always display key fields.
			defns = list.getFieldDefns();
			rows = list.retrieveVisibleRows();
			for (idxFld = 0; idxFld < defns.size(); idxFld++) {
				defn = defns.get(idxFld);
				if (defn.isVisible() && !defn.isKey()) {
					fieldSameInAllRows = true;
					fieldContainsDifference = false;
					//if (ignoreFields.contains(defn.getLabel())) {
					//	defn.setVisible(false);
					//}
					for (idxRow = 0; idxRow < rows.size() && defn.isVisible() && !fieldContainsDifference; idxRow++) {
						field = rows.get(idxRow).getField(defn.getLabel());
						if (field == null) {
							value = ""; //$NON-NLS-1$
						}
						else {	
							value = field.getValue();
						}
						if (idxRow == 0) {
							prevValue = value;
						}
						if (!(value.equals(prevValue))) {
							fieldContainsDifference = true;
						}
						prevValue = value;
					}
					if (fieldSameInAllRows && !fieldContainsDifference) {
						defn.setHiddenByUser(true);	//This is a work-around to field.setVisible(false) that should maintain current function. If false was applied to server_hidden that would hide the field in the source list until refresh...at least using the user setting this can be restored
					}
				}
			}
		}
		
		return list;
	}
	
	public void clearSources() {
		
		sourceIds.clear();
	}
	
	public void addSource(ComparatorSource source) {
	
		addSource(source.getId());
		
		defaultFieldSettings();
	}
	
	public void addSource(int id) {
		
		sourceIds.add(id);
	}
	
	public void addSources(ComparatorSource[] sources) {
		
		int idx;
		
		for (idx = 0; idx < sources.length; idx++) {
			addSource(sources[idx].getId());
		}
		
		defaultFieldSettings();
	}

	public void setSources(ComparatorSource[] sources) {
		
		int idx;
		
		clearSources();
		for (idx = 0; idx < sources.length; idx++) {
			addSource(sources[idx].getId());
		}
		
		defaultFieldSettings();
	}
	
	public int getSourceCount() {
		
		return sourceIds.size();
	}
	
	private void defaultFieldSettings() {
		
		int idxSce;
		int idxLbl;
		int idxDefn;
		DisplayList list;
		ArrayList<ListFieldDefn> defns;
		boolean fnd;
		ListFieldDefn defn;
		
		ignoreFields.clear();
		hideFields.clear();
		labels = getUniqueLabels();
		initialiseSourceFields(labels.length, sourceIds.size());
		
		//Fields which don't exist in each data source are defaulted to ignore and hide, unless they are key fields.
		for (idxLbl = 0; idxLbl < labels.length; idxLbl++) {
			for (idxSce = 0; idxSce < sourceIds.size(); idxSce++) {
				list = (DisplayList) instance.getComparatorSource(sourceIds.get(idxSce)).getData(instance);
				defns = list.getFieldDefns();
				fnd = false;
				for (idxDefn = 0; idxDefn < defns.size(); idxDefn++) {
					defn = defns.get(idxDefn);
					//Default ignore and hide if defined (usually for instant comparisons).
					if (defn.ignoreInCompare()) {
						if (!(ignoreFields.contains(defn.getLabel()))) {
							ignoreFields.add(defn.getLabel());
						}
					}
					if (defn.hideInCompare()) {
						if (!(hideFields.contains(defn.getLabel()))) {
							hideFields.add(defn.getLabel());
						}
					}
					
					if (defn.isVisible() && defn.getLabel().equalsIgnoreCase(labels[idxLbl])) {
						fnd = true;
						break;
					}
				}
				if (!fnd) {
					if (!(ignoreFields.contains(labels[idxLbl]))) {
						ignoreFields.add(labels[idxLbl]);
					}
					if (!(hideFields.contains(labels[idxLbl]))) {
						hideFields.add(labels[idxLbl]);
					}
					sourceFields[idxLbl][idxSce] = ""; //$NON-NLS-1$
				}
				else {
					sourceFields[idxLbl][idxSce] = labels[idxLbl];
				}
			}
		}
	}
	
	public ArrayList<String> getSourceLabels(int sourceIdx) {
		
		ArrayList<String> sourceLabels = new ArrayList<String>();
		DisplayList list;
		ArrayList<ListFieldDefn> defns;
		int idxFld;
		
		if (sourceIdx < sourceIds.size()) {  //The list of comparators can change if the user closes/refreshes displays.
			if (instance.getComparatorSource(sourceIds.get(sourceIdx)) != null) {
				list = (DisplayList) instance.getComparatorSource(sourceIds.get(sourceIdx)).getData(instance);
				defns = list.getFieldDefns();
				for (idxFld = 0; idxFld < defns.size(); idxFld++) {
					if (defns.get(idxFld).isVisible()) {
						sourceLabels.add(defns.get(idxFld).getLabel());
					}
				}
			}
		}
		
		return sourceLabels;
	}
	
	private String[] getUniqueLabels() {
		
		int idxSce;
		int idxLbl;
		int idxKey;
		DisplayList origList;
		ArrayList<ListFieldDefn> defns;
		ArrayList<String> labelsList = new ArrayList<String>();
		//String[] uniqueLabels;
		ListFieldDefn defn;
		
		/*
		 * Intentionally keeping the fields in the order they arrived from the data sources, so that the user is roughly familiar with field positions (especially key fields).
		 * 
		 * Maintain the key, ignore and hide fields.
		 * If not already defined, default key fields to the key fields on the underlying sources.
		 * If already defined, remove any fields that aren't in the current sources.
		 */
		for (idxSce = 0; idxSce < sourceIds.size(); idxSce++) {
			origList = (DisplayList) instance.getComparatorSource(sourceIds.get(idxSce)).getData(instance);
			defns = origList.getFieldDefns();
			for (idxLbl = 0; idxLbl < defns.size(); idxLbl++) {
				defn = defns.get(idxLbl);
				if (defn.isVisible() && !(labelsList.contains(defn.getLabel()))) {
					labelsList.add(defn.getLabel());
					if (defn.isKey() && !(keyFields.contains(defn.getLabel()))) {
						keyFields.add(defn.getLabel());
					}
				}
			}
		}
		
		for (idxKey = 0; idxKey < keyFields.size(); idxKey++) {
			if (!(labelsList.contains(keyFields.get(idxKey)))) {
				keyFields.remove(idxKey);
			}
		}
		for (idxLbl = 0; idxLbl < ignoreFields.size(); idxLbl++) {
			if (!(labelsList.contains(ignoreFields.get(idxLbl)))) {
				ignoreFields.remove(idxLbl);
			}
		}
		for (idxLbl = 0; idxLbl < hideFields.size(); idxLbl++) {
			if (!(labelsList.contains(hideFields.get(idxLbl)))) {
				hideFields.remove(idxLbl);
			}
		}
		
		labels = new String[labelsList.size()];
		for (idxLbl = 0; idxLbl < labelsList.size(); idxLbl++) {
			labels[idxLbl] = labelsList.get(idxLbl);
		}
		return labels;
	}
	
	public String getKeySequence(String label) {
		
		int idxKey;
		
		for (idxKey = 0; idxKey < keyFields.size(); idxKey++) {
			if (keyFields.get(idxKey).equalsIgnoreCase(label)) {
				return Integer.toString(idxKey + 1);
			}
		}
		
		return ""; //$NON-NLS-1$
	}
	
	public boolean isIgnoreField(String label) {
		
		int idxFld;
		
		for (idxFld = 0; idxFld < ignoreFields.size(); idxFld++) {
			if (ignoreFields.get(idxFld).equalsIgnoreCase(label)) {
				return true;
			}
		}
		
		return false;
	}

	public boolean isHideField(String label) {
	
		int idxFld;
		
		for (idxFld = 0; idxFld < hideFields.size(); idxFld++) {
			if (hideFields.get(idxFld).equalsIgnoreCase(label)) {
				return true;
			}
		}
		
		return false;
	}
	
	public ArrayList<ComparatorSource> getCurrentSources() {
		
		ArrayList<ComparatorSource> allSources = new ArrayList<ComparatorSource>();
		ArrayList<ComparatorSource> currentSources = new ArrayList<ComparatorSource>();
		ComparatorSource sce;
		int idxAll;
		
		allSources.addAll(instance.getOpenComparatorSources());
		allSources.addAll(instance.getSavedDisplayComparatorSources());
		
		for (idxAll = 0; idxAll < allSources.size(); idxAll++) {
			sce = allSources.get(idxAll);
			if (sourceIds.contains(sce.getId())) {
				currentSources.add(sce);
			}
		}
		
		Collections.sort(currentSources);  //Must be sorted to keep in synch with ComparatorConfigurationDialog.
		
		return currentSources;
	}
	
	public String editAndSet(ArrayList<ComparatorField> fields) {
		
		String msg = ""; //$NON-NLS-1$
		int idxLbl;
		int idxDup;
		int idxSce;
		int idxCnt;
		ComparatorField fld;
		boolean fnd;
		ComparatorSource source;
		
		for (idxLbl = 0; idxLbl < fields.size() && msg.length() == 0; idxLbl++) {
			fld = fields.get(idxLbl);
			for (idxDup = (idxLbl + 1); idxDup < fields.size(); idxDup++) {
				if (fields.get(idxDup).getLabel().equalsIgnoreCase(fld.getLabel())) {
					msg = Messages.getString("ComparatorHelper_5") + fld.getLabel() + Messages.getString("ComparatorHelper_6") + (idxLbl + 1) + Messages.getString("ComparatorHelper_7") + (idxDup + 1) + "."; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
					break;
				}
			}
		}
		
		if (msg.length() == 0) {
			for (idxLbl = 0; idxLbl < fields.size(); idxLbl++) {
				fld = fields.get(idxLbl);		
				if (fld.getKeySequence().trim().length() > 0) {
					if (fld.isHidden() || fld.isIgnored()) {
						msg = Messages.getString("ComparatorHelper_9") + fld.getLabel() + "."; //$NON-NLS-1$ //$NON-NLS-2$
						break;
					}
				}
			}
		}
		
		if (msg.length() == 0) {
			for (idxLbl = 0; idxLbl < fields.size(); idxLbl++) {
				fld = fields.get(idxLbl);		
				if (fld.getKeySequence().trim().length() > 0) {
					if (!((new NumberEditor(fld.getKeySequence())).isValidInteger(12))) {
						msg = Messages.getString("ComparatorHelper_11") + fld.getLabel() + "."; //$NON-NLS-1$ //$NON-NLS-2$
						break;
					}
				}
			}
		}
		
		if (msg.length() == 0) {
			clearHideFields();
			clearIgnoreFields();
			initialiseSourceFields(fields.size(), sourceIds.size());
			labels = new String[fields.size()];
			setKeyFields(fields);
			for (idxLbl = 0; idxLbl < fields.size(); idxLbl++) {
				fld = fields.get(idxLbl);
				if (fld.isHidden()) {
					hideFields.add(fld.getLabel());
				}
				if (fld.isIgnored()) {
					ignoreFields.add(fld.getLabel());
				}
				labels[idxLbl] = fld.getLabel();
				for (idxSce = 0; idxSce < sourceIds.size(); idxSce++) {
					fnd = false;
					for (idxCnt = 0; idxCnt < fld.getSources().size(); idxCnt++) {
						source = fld.getSources().get(idxCnt);
						if (source.getId() == sourceIds.get(idxSce)) {
							fnd = true;
							sourceFields[idxLbl][idxSce] = fld.getSourceFields().get(idxCnt);
							break;
						}
					}
					if (!fnd) {
						sourceFields[idxLbl][idxSce] = ""; //$NON-NLS-1$
					}
				}
			}
			
			saveConfiguration(fields);
		}
		
		return msg;
	}
	
	private void saveConfiguration(ArrayList<ComparatorField> fields) {
		
		DisplayList list = null;
		ComparatorConfiguration config = null;
		JAXBContext 	xmlContext;
		Marshaller 	m;
		File file;
		FileOutputStream out;
		
		if (sourceIds.size() > 0) {
			list = (DisplayList) instance.getComparatorSource(sourceIds.get(0)).getData(instance);
			if (list != null) {
				config = new ComparatorConfiguration(list.getContext());
				config.setConfiguration(this, fields);
				
				try {
					file = getSavedConfigFile(list);
					out = new FileOutputStream(file.getAbsolutePath());
					
					xmlContext = JAXBContext.newInstance(ComparatorConfiguration.class);
					m = xmlContext.createMarshaller();
					m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
					m.marshal(config, out);
					out.close();
				}
				catch (Exception e) {
					instance.addConsoleMessage(e);
				}
			}
		}
	}
	
	private void restoreSavedConfiguration() {
		
		File file;
		DisplayList list;
		ComparatorConfiguration config;
		FileInputStream in;
		JAXBContext xmlContext;
		Unmarshaller u;
		
		list = (DisplayList) instance.getComparatorSource(sourceIds.get(0)).getData(instance);
		
		try {
			file = getSavedConfigFile(list);
			in = new FileInputStream(file);

			xmlContext = JAXBContext.newInstance(ComparatorConfiguration.class);
			u = xmlContext.createUnmarshaller();
			config = (ComparatorConfiguration) u.unmarshal(in);
			in.close();
			
			config.restoreConfiguration(this);
		} 
		catch (Throwable e) {
			//instance.addConsoleMessage(e);  //If can't open a saved configuration, doesn't matter as the metadata defaults are used.
		}
	}

	private File getSavedConfigFile(DisplayList list) {
		
		return new File(instance.getWorkingFolder(), savedConfigurationPfx + list.getContext() + savedConfigurationSfx);
	}
	
	private void clearIgnoreFields() {
		
		ignoreFields.clear();
	}
	
	private void clearHideFields() {
		
		hideFields.clear();
	}
	
	public void setIgnoreField(String label) {
		
		ignoreFields.add(label);
	}
	
	public void setHideField(String label) {
		
		hideFields.add(label);
	}

	public String[][] getSourceFields() {
		
		return sourceFields;
	}

	public String[] getLabels() {
		
		return labels;
	}
	
	private void initialiseSourceFields(int lblCnt, int sceCnt) {
		
		int idxLbl;
		int idxSce;
		
		sourceFields = new String[labels.length][sourceIds.size()];
		for (idxLbl = 0; idxLbl < lblCnt; idxLbl++) {
			for (idxSce = 0; idxSce < sceCnt; idxSce++) {
				sourceFields[idxLbl][idxSce] = ""; //$NON-NLS-1$
			}
		}
	}
	
	private void setKeyFields(ArrayList<ComparatorField> fields) {
		
		int idxFld;
		int idxKey;
		ArrayList<Integer> sequences = new ArrayList<Integer>();
		ComparatorField field;
		boolean placed;
		int sequence;
		
		keyFields.clear();
		
		//Need to add the key fields in sequence order.
		for (idxFld = 0; idxFld < fields.size(); idxFld++) {
			field = fields.get(idxFld);
			if (field.getKeySequence().trim().length() > 0) {
				sequence = Integer.parseInt(field.getKeySequence());
				placed = false;
				if (keyFields.size() > 0) {
					for (idxKey = 0; idxKey < keyFields.size(); idxKey++) {
						if (sequences.get(idxKey) > sequence) {
							keyFields.add(idxKey, field.getLabel());
							sequences.add(idxKey, sequence);
							placed = true;
							break;
						}
					}
				}
				if (!placed) {
					keyFields.add(field.getLabel());
					sequences.add(sequence);
				}
			}
		}
	}

	public void setShowAllKeys(boolean showAllKeys) {
		
		this.showAllKeys = showAllKeys;
	}

	public void setShowMatchedKeys(boolean showMatchedKeys) {
		
		this.showMatchedKeys = showMatchedKeys;
	}

	public void setShowUnmatchedKeys(boolean showUnmatchedKeys) {
		
		this.showUnmatchedKeys = showUnmatchedKeys;
	}

	public boolean showingAllKeys() {
		
		return showAllKeys;
	}

	public boolean showingMatchedKeys() {
		
		return showMatchedKeys;
	}

	public boolean showingUnmatchedKeys() {
		
		return showUnmatchedKeys;
	}
	
	public String getSourcesAsTooltipText() {
		
		int idxSce;
		String text = ""; //$NON-NLS-1$
		
		for (idxSce = 0; idxSce < sourceIds.size(); idxSce++) {
			if (idxSce == 0) {
				text = Messages.getString("ComparatorHelper_16"); //$NON-NLS-1$
			}
			if (instance.getComparatorSource(sourceIds.get(idxSce)) != null) {  //The current source may have been removed.
				text = text + "\r\n" + (idxSce + 1) + " = " + instance.getComparatorSource(sourceIds.get(idxSce)).getTitle(); //$NON-NLS-1$ //$NON-NLS-2$
			}
		}
		
		return text;
	}
	
	public int findDifference(DisplayList list, boolean forward, int startRow) {
		
		int idxRow = startRow;
		int idxFld;
		boolean found = false;
		ListRow row;
		ArrayList<ListRow> rows = list.getVisibleRows();
		ArrayList<ListField> fields;
		
		if (list.getVisibleRows().size() > 0) {
			while (!found) {
				if ((forward && idxRow >= list.getVisibleRowCount()) || (!forward && idxRow < 0)) {
					break;
				}
				row = rows.get(idxRow);
				fields = row.getFields();
				for (idxFld = 0; idxFld < fields.size(); idxFld++) {
					if (fields.get(idxFld).isVisible()) {
						if (fields.get(idxFld).hasDifferentValue()) {
							found = true;
							break;
						}
					}
				}
				
				if (!found) {
					if (forward) {
						idxRow++;
					}
					else {
						idxRow--;
					}
				}
			}
		}
		
		if (found) {
			return idxRow;
		}
		else {
			return -1;
		}
	}

	public ArrayList<String> getKeyFields() {
		
		return keyFields;
	}

	public ArrayList<String> getIgnoreFields() {
		
		return ignoreFields;
	}

	public ArrayList<String> getHideFields() {
		
		return hideFields;
	}

	public void restoreKeyFields(ArrayList<String> keyFields) {
		this.keyFields = keyFields;
	}

	public void restoreIgnoreFields(ArrayList<String> ignoreFields) {
		this.ignoreFields = ignoreFields;
	}

	public void restoreHideFields(ArrayList<String> hideFields) {
		this.hideFields = hideFields;
	}
}